
% ============================================================
% Function that computes and plots horizontal force from 3D
% CFD-FNV load model fra 2D CFD outout in OF
% ============================================================
% Master's thesis: Modelling of Ringing Loads on Offshore Wind
% Turbine monopiles
% Department of Marine Technology, NTNU
% Spring 2020
% ============================================================
% Modified: Vilde Mathisen, 25.05.20
% ============================================================


clear all
clf

%% Parameters
ha = 7.83;
ka = 0.19;
steep_1 = 0.04;
steep = 1/steep_1;
nharm = 5;

if ka == 0.105
    T = 2.021;
elseif ka == 0.127
    T = 1.732;
elseif ka == 0.163
    T = 1.443;
elseif ka == 0.388
    T = 0.866;
elseif ka == 0.290
    T = 1.010;
elseif ka == 0.229
    T = 1.155;
elseif ka == 0.190
    T = 1.299;
elseif ka == 0.142
    T = 1.588;
elseif ka == 0.115
    T = 1.876;
elseif ka ==0.097
    T = 2.165;
elseif ka == 0.090
    T = 2.309;
else 
    disp('Invalid ka number');
end

%% Constant
scale = 48;
d = 6.9/scale;
aa = d/2;
h = ha*aa;
hh1 = h;
rho = 1000;
g = 9.81;

%% Probes properties
n = 100;                %number of radial probes
Nc = 160;                 %integration point around cylinder
probeline = [1:Nc];
[dl theta] = probes();
delta_s = (pi*d)/Nc;

pth0 = ['Stokes_waves/ha',num2str(ha),'/ka',num2str(ka),'/steep',num2str(steep_1),'/']; 

%% Strips
M = 10;                   %no. of strips
m = [1:1:M];              %strip vector

for i = 1:length(m)
   [time F_CFD_strip(i,:) z(i) delta_z(i)] = computeStripForce(pth0,i); 
   F_CFD_z(i,:) = F_CFD_strip(i,:)*delta_z(i);
end

F_CFD = sum(F_CFD_z);
dt    = time(2)-time(1);
[F_FNV zeta] = modifiedFNV(rho,aa,T,h,steep,time,nharm,hh1,g);

for i = 1:length(m)
   F_CFD_zeta_strip(i,:) = zetaForce(zeta,time,F_CFD_z(i,:),z(i),delta_z(i)); 
end

for i = 1:length(m)
    [F_w_strip(i,:) time] = readIntegratedForce(pth0,i);
    F_w_z(i,:) = F_w_strip(i,:)*delta_z(i);
end

F_w = sum(F_w_z);

F_CFD_zeta = sum(F_CFD_zeta_strip);

Fx = F_CFD_zeta + F_FNV + F_w;

[FA FPSI FDrag] = originalFNV(rho,aa,T,h,steep,time,nharm,h,g);
    
    
set(gcf, 'Position',  [100, 100, 700, 700])
    
    for iharm = 1:nharm
      flow  = (iharm - 0.4)/T;
      fhigh = (iharm + 0.4)/T;
      
      y   = (FA + FPSI + FDrag)'/(rho*g*aa^3);
      ybp = bpass2(y,dt,flow,fhigh,0,0,'',0.0002);
      yFx = bpass2(Fx',dt,flow,fhigh,0,0,'',0.0002);
      yFx_CFD = bpass2((F_CFD_zeta + F_FNV)',dt,flow,fhigh,0,0,'',0.0002);
      ampFNV(iharm) = max(ybp);
      ampFx(iharm) = max(yFx/(rho*g*aa^3));
      ampFx_CFD(iharm) = max(yFx_CFD/(rho*g*aa^3));
      
      figure(1)

      xlim([0 8])
      subplot(4,1,1)
      title(['$ka = ' num2str(ka) '$, $H_{1}/\lambda = 1/' num2str(steep) '$, $T = ' num2str(T) '$'],'Interpreter','latex','FontSize',16);
      plot(time/T,y,'LineWidth',0.7,'Color',[0.8 0.001 0])
      hold on, grid on
      plot(time/T,Fx/(rho*g*aa^3),'Linewidth',0.7,'Color',[0 0.4 0.8])
      plot(time/T,(F_CFD_zeta + F_FNV)/(rho*g*aa^3),'--','Linewidth',0.7,'Color',[0 0.4 0.8])
      ylabel(['$\frac{F_{x}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
      legend('$F_{FNV}$','$F_{3D-CFD-FNV}$','$F_{CFD-FNV}$','Location','northeast','Interpreter','latex','FontSize',13);
      if (iharm <= 3)
        subplot(4,1,iharm+1)
        plot(time/T,ybp,'LineWidth',0.7,'Color',[0.8 0.001 0])
        hold on, grid on
        plot(time/T,yFx/(rho*g*aa^3),'Linewidth',0.7,'Color',[0 0.4 0.8])
        plot(time/T,yFx_CFD/(rho*g*aa^3),'--','Linewidth',0.7,'Color',[0 0.4 0.8])
        if (iharm >= 2)
        ylabel(['$\frac{F_{x}^{(' num2str(iharm,'%i') '\omega)}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
          else 
        ylabel(['$\frac{F_{x}^{(\omega)}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
        end
         if(iharm ==3)
            xlabel('t/T','Interpreter','latex','FontSize',17);
         end


      yCFD = bpass2(F_CFD_zeta',dt,flow,fhigh,0,0,'',0.0002);
      yFNV = bpass2(F_FNV',dt,flow,fhigh,0,0,'',0.0002);
      yFw = bpass2(F_w',dt,flow,fhigh,0,0,'',0.0002);
      ampCFD(iharm) = max(yCFD/(rho*g*aa^3));
      ampFNV(iharm) = max(yFNV/(rho*g*aa^3));
      ampFw(iharm) = max(yFw/(rho*g*aa^3));

      
      figure(2)
      subplot(4,1,1)
      title(['$ka = ' num2str(ka) '$, $H_{1}/\lambda = 1/' num2str(steep) '$, $T = ' num2str(T) '$'],'Interpreter','latex','FontSize',16);
      plot(time/T,F_CFD/(rho*g*aa^3),'LineWidth',0.7,'Color',[0.1 0.5 0.9])
      hold on, grid on
      plot(time/T,F_FNV/(rho*g*aa^3),'Linewidth',0.7,'Color',[1 0.5 0])
      plot(time/T,F_w/(rho*g*aa^3),'Linewidth',0.7,'Color',[0.1 0.4 0.1])
            xlim([0 8])
      ylabel(['$\frac{F_{x}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
      legend('$F^{CFD}$','$F^{FNV}$','$F^{w}$','Location','northeast','Interpreter','latex','FontSize',13);
      if (iharm <= 3)
        subplot(4,1,iharm+1)
        plot(time/T,yCFD/(rho*g*aa^3),'r','LineWidth',0.7,'Color',[0.1 0.5 0.9])
        hold on, grid on
        plot(time/T,yFNV/(rho*g*aa^3),'g','Linewidth',0.7,'Color',[1 0.5 0])
        plot(time/T,yFw/(rho*g*aa^3),'b','Linewidth',0.7,'Color',[0.1 0.4 0.1])
         xlim([0 8])
        if (iharm >= 2)
        ylabel(['$\frac{F_{x}^{(' num2str(iharm,'%i') '\omega)}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
          else 
        ylabel(['$\frac{F_{x}^{(\omega)}}{\rho g a^{3}}$'],'Interpreter','latex','FontSize',25);
        end
         if(iharm ==3)
            xlabel('t/T','Interpreter','latex','FontSize',17);
         end

     end


     end
    end
    
    
%% Functions

    
function [F_CFD_zeta] = zetaForce(zeta,time,F_CFD,z_top,delta_z)
  mid_strip = z_top - delta_z/2;
  for i = 1:length(time)
      if zeta(i) <= mid_strip
         F_CFD_zeta(i)  = 0;
      elseif zeta(i) > mid_strip
         F_CFD_zeta(i) = F_CFD(i);
     end
  end
 end  


function [time F_CFD z_2 delta_z] = computeStripForce(pth0,m)
    pth_strip = [pth0 'Strips/strip',num2str(m),'/'];
    [time F2D] = read2DForce(pth_strip);
    [z_1 z_2 delta_z] = readDeltaZ(pth_strip);
    F_CFD = F2D;
end


%% Functions for reading files

function [F_int time] = readIntegratedForce(pth,m);

pth_strip = [pth 'Strips/strip',num2str(m),'/'];

fname = [pth_strip 'Fw.txt'];
fid = fopen(fname, 'r');
if (fid == -1)
    error('FFL:fopen', 'Unable to open: %s', fname);
end

dataRaw = textscan(fid, '%f %f ');
data = cell2mat(dataRaw);
time = data(:,1);
F_int = data(:,2);

fclose(fid);

end

function [time F2D] = read2DForce(pth);

fname = [pth '2D_force.txt'];
fid = fopen(fname, 'r');
if (fid == -1)
    error('FFL:fopen', 'Unable to open: %s', fname);
end

dataRaw = textscan(fid, '%f %f');
data = cell2mat(dataRaw);
time = data(:,1);
F2D = data(:,2);

fclose(fid);

end



function [z_1 z_2 delta_z] = readDeltaZ(pth);

fname = [pth 'delta_z.txt'];
fid = fopen(fname, 'r');
if (fid == -1)
    error('FFL:fopen', 'Unable to open: %s', fname);
end

dataRaw = textscan(fid, '%f %f %f');
data = cell2mat(dataRaw);
delta_z = data(1);
z_1 = data(2);
z_2 = data(3);


fclose(fid);

end

%% FNV

function [FA FPSI FDrag] = originalFNV(rho,aa,T,h,steep,time,nharm,hh1,g);


a11 = rho*pi*aa^2;
fk  = rho*pi*aa^2;

  om = 2*pi/T;
  
  %% Calculate theoretical Stokes 5 kinematics from Skjelbreia:
  kw1 = findk(T,h);

  za1 = pi/(kw1*steep);
  H   = 2*za1;%Hm(iper);
  k = kw1;
  za = za1;
  
  %% z and time:
  xpt   = 0;
  x     = xpt;
  
  Nz = 1000;
  dz    = h/Nz;
  zvec0  = [-Nz:0]'*dz;
  
   dt    = time(2)-time(1);
   Nt    = length(time);
%    
%      NTPER = 240;
%   NPER  = 8;
%   dt    = T/NTPER;
%   time  = [1:NTPER*NPER]'*dt;
%   Nt    = length(time);

  %% Compute the constants; A, B and C's:
  [A11 A13 A15 A22 A24 A33 A35 A44 A55 B22 B24 B33 B35 B44 B55 C1 C2 C3 C4] ...
    = computeSkjelbreiaConstants(k,h);
  
  %% Kinematics for z <= 0:
  th  = k*x - om*time';
  kz  = k*(h + zvec0);
  kh  = k*h;
  lam = k*za;
  
  dum1 = om/k^2*(lam  *A11 + lam^3*A13 + lam^5*A15);
  dum2 = om/k^2*(lam^2*A22 + lam^4*A24          );
  dum3 = om/k^2*(lam^3*A33 + lam^5*A35          );
  dum4 = om/k^2*(lam^4*A44                      );
  dum5 = om/k^2*(lam^5*A55                      );
  
  phi  = (dum1*cosh(kz)*sin(th)     + ...
    dum2*cosh(2*kz)*sin(2*th) + ...
    dum3*cosh(3*kz)*sin(3*th) + ...
    dum4*cosh(4*kz)*sin(4*th) + ...
    dum5*cosh(5*kz)*sin(5*th));
  
  u  = k*(dum1*cosh(kz)*cos(th)     + ...
    2*dum2*cosh(2*kz)*cos(2*th) + ...
    3*dum3*cosh(3*kz)*cos(3*th) + ...
    4*dum4*cosh(4*kz)*cos(4*th) + ...
    5*dum5*cosh(5*kz)*cos(5*th));
  
  w  = k*(dum1*sinh(kz)*sin(th)     + ...
    2*dum2*sinh(2*kz)*sin(2*th) + ...
    3*dum3*sinh(3*kz)*sin(3*th) + ...
    4*dum4*sinh(4*kz)*sin(4*th) + ...
    5*dum5*sinh(5*kz)*sin(5*th));
  
  ut =  k*om*(dum1*cosh(kz)*sin(th)  + ...
    4*dum2*cosh(2*kz)*sin(2*th) + ...
    9*dum3*cosh(3*kz)*sin(3*th) + ...
    16*dum4*cosh(4*kz)*sin(4*th) + ...
    25*dum5*cosh(5*kz)*sin(5*th));
  
  ux = -k^2*(dum1*cosh(kz)*sin(th)  + ...
    4*dum2*cosh(2*kz)*sin(2*th) + ...
    9*dum3*cosh(3*kz)*sin(3*th) + ...
    16*dum4*cosh(4*kz)*sin(4*th) + ...
    25*dum5*cosh(5*kz)*sin(5*th));
  
  uz = k^2*(dum1*sinh(kz)*cos(th)   + ...
    4*dum2*sinh(2*kz)*cos(2*th) + ...
    9*dum3*sinh(3*kz)*cos(3*th) + ...
    16*dum4*sinh(4*kz)*cos(4*th) + ...
    25*dum5*sinh(5*kz)*cos(5*th));
  
  zeta = 1/k*((lam                  )*cos(th) + ...
    (lam^2*B22 + lam^4*B24)*cos(2*th) + ...
    (lam^3*B33 + lam^5*B35)*cos(3*th) + ...
    (lam^4*B44            )*cos(4*th) + ...
    (lam^5*B55            )*cos(5*th));

  %% Harmonics of zeta:
  Zeta(1) = 1/k*lam;
  Zeta(2) = 1/k*(lam^2*B22 + lam^4*B24);
  Zeta(3) = 1/k*(lam^3*B33 + lam^5*B35);
  Zeta(4) = 1/k*(lam^4*B44);
  Zeta(5) = 1/k*(lam^5*B55);
  
  %% For zeta > 0:
  % th  = k*x - om*time
  phi1  =  dum1*cosh(kz)*sin(th);
  phi2  =  dum2*cosh(2*kz)*sin(2*th);
  phi3  =  dum3*cosh(3*kz)*sin(3*th);
  phi4  =  dum4*cosh(4*kz)*sin(4*th);
  phi5  =  dum5*cosh(5*kz)*sin(5*th);
  
  cphi1  =  'dum1*cosh(kz)*sin(th);';
  cphi2  =  'dum2*cosh(2*kz)*sin(2*th);';
  cphi3  =  'dum3*cosh(3*kz)*sin(3*th);';
  cphi4  =  'dum4*cosh(4*kz)*sin(4*th);';
  cphi5  =  'dum5*cosh(5*kz)*sin(5*th);';

  %% Rules:
  % d/dt : fac = om; sin  -> -cos;  cos  ->  sin;
  % d/dx : fac = k;  sin  ->  cos;  cos  -> -sin;
  % d/dz : fac = k;  sinh ->  cosh; cosh -> sinh;
  
  kinvec   = {'u','w','ut','ux','uz'};
  irulevec = [ 2   3   2    2    2  ; ...
               0   0   1    2    3   ];
  nkin    = length(kinvec);
  
  rule0  = {'',''};               %  -
  rule1  = {'sin(' ,'cos('} ;     % d/dt
  rule2  = {'sin(' ,'cos('} ;     % d/dx
  rule3  = {'sinh(','cosh('};     % d/dz
  signs  = [  1        1
             -1        1    ; ...
              1       -1    ; ...
              1        1    ];
  facs   = { '1'        '1'  ; ...
             'om'      'om'  ; ...
             'k'        'k'  ; ...
             'k'        'k'  };
  
  %% "Compute" {'u','w','ut','ux','uz'}:
  KIN = zeros(Nz+1,Nt,nkin,nharm);
  for ikin = 1:nkin
    kin = char(kinvec(ikin));
    
    for iharm = 1:nharm
      charm = num2str(iharm);
      eval(['a = cphi' charm ';'])
      
      pm0 = 1;
      for ir = 1:2
        irule = irulevec(ir,ikin);
        eval(['rule = rule' num2str(irule) ';'])
        a1 = char(rule{1});
        a2 = char(rule{2});
        
        ib = strfind(a,a1);
        if (length(ib > 0))
          c    = a1;
          d    = a2;
          pm   = signs(irule+1,1);
          fac  = char(facs(irule+1,1));
        else
          ib = strfind(a,a2);
          c    = a2;
          d    = a1;
          pm   = signs(irule+1,2);
          fac  = char(facs(irule+1,2));
        end
        nr   = length(c);
        ivec = [ib : ib + nr - 1];
        if (length(ivec) > 0)
          a(ivec) = [];
          a = [a(1:ib-1) d a(ib:end)];
        end
        if (irule>0)
          a = [ num2str(iharm) '*' fac '*' a ];
          pm  = pm*pm0;
          pm0 = pm;
        end
      end
      if (pm == -1)
        a = [ num2str(pm) '*' a ];
      end
      %disp(['For z <= 0. ' kin ' :   ' a])
      A{ikin,iharm} = a;
      eval(['dum = ' a ';']);
      KIN(:,:,ikin,iharm) = dum;
    end
  end

   %% Taylor exapansion of {'u','w','ut','ux','uz'}:
  irule = 3;
  rule  = rule3;
  Nder  = 4;
  zm  = max(zeta);
  Nz1 = 120;
  dz  = zm/Nz1;
  z   = [0:Nz1]'*dz;
  K   = zeros(Nz1+1,Nt,nkin);
  
  for ikin = 1:nkin
    kin = char(kinvec(ikin));
    
    %% Zeroth derivative:
    ider = 0;
    for iharm = 1:nharm
      a    = A{ikin,iharm};
      ib   = strfind(a,'kz');
      nr   = length('kz');
      ivec = [ib : ib + nr - 1];
      a(ivec) = [];
      a    = [a(1:ib-1) 'kh' a(ib:end)];
      A0{ikin,iharm,ider+1} = a;
    end
    %['Text 1: ' kin ],A0(ikin,1,:)
    
    %% Derivatives 1-4:
    for ider = 1:Nder
      for iharm = 1:nharm-ider
        a = A0{ikin,iharm,ider};
        
        pm0 = 1;
        a1 = char(rule{1});
        a2 = char(rule{2});
        
        ib = strfind(a,a1);
        if (length(ib > 0))
          c    = a1;
          d    = a2;
          pm   = signs(irule+1,1);
          fac  = char(facs(irule+1,1));
        else
          ib = strfind(a,a2);
          c    = a2;
          d    = a1;
          pm   = signs(irule+1,2);
          fac  = char(facs(irule+1,2));
        end
        nr   = length(c);
        ivec = [ib : ib + nr - 1];
        if (length(ivec) > 0)
          a(ivec) = [];
          a = [a(1:ib-1) d a(ib:end)];
        end
        a = [ num2str(iharm) '*' fac '*' a ];
        pm  = pm*pm0;
        pm0 = pm;
        a = [ num2str(pm) '*' a ];
        
        A0{ikin,iharm,ider+1} = a;
      end
   end
    
    %% Sum up contributions for each z:
    f = zeros(Nz1+1,Nt);
    
    ider = 0;
    for iharm = 1:nharm
      a = A0{ikin,iharm,ider+1};
      eval(['f = f + ones(Nz1+1,1)*' a ])
    end

    for ider = 1:Nder
      for iharm = 1:nharm-ider
        a = A0{ikin,iharm,ider+1};
        str = ['f = f + z.^' num2str(ider) '*' a(1:end-1) '/gamma(' num2str(ider+1) ');'];
        eval(str)
      end
    end
    K(:,:,ikin) = f;
  end
  
  %% Calculate FNV forces:
  u0 = squeeze(K(:,:,1));
  w0 = squeeze(K(:,:,2));
  ut0 = squeeze(K(:,:,3));
  ux0 = squeeze(K(:,:,4));
  uz0 = squeeze(K(:,:,5));
  
   %% Calculate KC5:
  iKCvar = 0;
  if (iKCvar == 1)
    umax = zeros(Nz + Nz1 + 2,1);
    for it = 1:Nt
      if (zeta(it) >=0 )
        ivec0 = find(z <= zeta(it));
        ivec  = [1:Nz+1]';
        uu    = [u(:,it) ; u0(ivec0,it)];
      else
        ivec0 = [];
        ivec  = find(zvec0 <= zeta(it));
        uu    = u(ivec,it);
      end
      ivec1 = [ivec ; ivec0];
      umax(ivec1) = max(umax(ivec1),uu);
    end
    KC5z = umax*T/(2*aa);
  end
  
  tic
  %% Arrange kinematics per time step:
  for it = 1:Nt
    if (zeta(it) >=0 )
      ivec = find(z <= zeta(it));
      uu  = [u(:,it) ; u0(ivec,it)];
      ww  = [w(:,it) ; w0(ivec,it)];
      uut = [ut(:,it) ; ut0(ivec,it)];
      uux = [ux(:,it) ; ux0(ivec,it)];
      uuz = [uz(:,it) ; uz0(ivec,it)];
      zv  = [zvec0 ; z(ivec)];
    else
      ivec = find(zvec0 <= zeta(it));
      uu = u(ivec,it);
      ww = w(ivec,it);
      uut = ut(ivec,it);
      uux = ux(ivec,it);
      uuz = uz(ivec,it);
      zv  = [zvec0(ivec)];
    end

    dzv = diff(zv);
    zvm = 0.5*(zv(1:end-1) + zv(2:end)) + hh1;
    
    um  = 0.5*(uu(1:end-1) + uu(2:end)); 
    wm  = 0.5*(ww(1:end-1) + ww(2:end)); 
    utm = 0.5*(uut(1:end-1) + uut(2:end)); 
    uxm = 0.5*(uux(1:end-1) + uux(2:end)); 
    uzm = 0.5*(uuz(1:end-1) + uuz(2:end)); 
    
    %% KC number dependent a11:
    if (iKCvar == 1)
      [Cm, Cd] = curveFittedCmCd(KC5z);
      a11z     = (Cm - 1)*rho*pi*aa^2;
      a11zm    = 0.5*(a11z(2:end,:) + a11z(1:end-1,:));
      a11      = a11zm(1:length(um));
    end
    
    %% FA:
    nzt(it) = length(zv);
    
    dum1    = (a11 + fk).*(utm);
    dum2    =          fk.*(um.*uxm);
    dum3    = (a11 + fk).*(wm.*uzm);
    dum     = dum1 + dum2 + dum3;
    FA(it)  = sum(dum.*dzv);
    
    MA(it)  = sum(dum.*zvm.*dzv);
   

    %% FPSI:
    u1    =  KIN(end,it,1,1);
    du1dt =  KIN(end,it,3,1);

    FPSI = fk*4/g*u1.^2.*du1dt;
    
    MPSI = FPSI*hh1;

  end
  
    %% KC dependent drag:
  if (iKCvar == 1)
    Cdm    = 0.5*(Cd(2:end) + Cd(1:end-1));
    whos Cd Cdm um
    dumd   = 0.5*rho*DD*Cdm(1:length(um)).*um.*abs(um);
    FDrag  = iKCvar*sum(dumd.*dzv);
    
    MDrag  = iKCvar*sum(dumd.*zvm.*dzv);
  else
    FDrag  = 0;
    MDrag  = 0;
  end
  
    
%% Plot harmonics
  %% Band-pass filter:
  for iharm = 1:nharm
    flow  = (iharm - 0.4)/T;
    fhigh = (iharm + 0.4)/T;
    nrep = 8;
    
    y   = repmat((FA + FPSI + FDrag),1,nrep)';
    ybp = bpass2(y,dt,flow,fhigh,0,0,'',0.0002);
    Fx(iharm) = max(ybp);

  end
  
end
  
  %% MODIFIED FNV
  
  function [F_FNV zeta] = modifiedFNV(rho,aa,T,h,steep,time,nharm,hh1,g);


a11 = rho*pi*aa^2;
fk  = rho*pi*aa^2;

  om = 2*pi/T;
  
  %% Calculate theoretical Stokes 5 kinematics from Skjelbreia:
  kw1 = findk(T,h);

  za1 = pi/(kw1*steep);
  H   = 2*za1;%Hm(iper);
  k = kw1;
  za = za1;
  
  %% z and time:
  xpt   = 0;
  x     = xpt;
  
  Nz = 1000;
  dz    = h/Nz;
  zvec0  = [-Nz:0]'*dz;
  
   dt    = time(2)-time(1);
   Nt    = length(time);

  %% Compute the constants; A, B and C's:
  [A11 A13 A15 A22 A24 A33 A35 A44 A55 B22 B24 B33 B35 B44 B55 C1 C2 C3 C4] ...
    = computeSkjelbreiaConstants(k,h);
  
  %% Kinematics for z <= 0:
  th  = k*x - om*time';
  kz  = k*(h + zvec0);
  kh  = k*h;
  lam = k*za;
  
  dum1 = om/k^2*(lam  *A11 + lam^3*A13 + lam^5*A15);
  dum2 = om/k^2*(lam^2*A22 + lam^4*A24          );
  dum3 = om/k^2*(lam^3*A33 + lam^5*A35          );
  dum4 = om/k^2*(lam^4*A44                      );
  dum5 = om/k^2*(lam^5*A55                      );
  
  phi  = (dum1*cosh(kz)*sin(th)     + ...
    dum2*cosh(2*kz)*sin(2*th) + ...
    dum3*cosh(3*kz)*sin(3*th) + ...
    dum4*cosh(4*kz)*sin(4*th) + ...
    dum5*cosh(5*kz)*sin(5*th));
  
  u  = k*(dum1*cosh(kz)*cos(th)     + ...
    2*dum2*cosh(2*kz)*cos(2*th) + ...
    3*dum3*cosh(3*kz)*cos(3*th) + ...
    4*dum4*cosh(4*kz)*cos(4*th) + ...
    5*dum5*cosh(5*kz)*cos(5*th));
  
  w  = k*(dum1*sinh(kz)*sin(th)     + ...
    2*dum2*sinh(2*kz)*sin(2*th) + ...
    3*dum3*sinh(3*kz)*sin(3*th) + ...
    4*dum4*sinh(4*kz)*sin(4*th) + ...
    5*dum5*sinh(5*kz)*sin(5*th));
  
  ut =  k*om*(dum1*cosh(kz)*sin(th)  + ...
    4*dum2*cosh(2*kz)*sin(2*th) + ...
    9*dum3*cosh(3*kz)*sin(3*th) + ...
    16*dum4*cosh(4*kz)*sin(4*th) + ...
    25*dum5*cosh(5*kz)*sin(5*th));
  
  ux = -k^2*(dum1*cosh(kz)*sin(th)  + ...
    4*dum2*cosh(2*kz)*sin(2*th) + ...
    9*dum3*cosh(3*kz)*sin(3*th) + ...
    16*dum4*cosh(4*kz)*sin(4*th) + ...
    25*dum5*cosh(5*kz)*sin(5*th));
  
  uz = k^2*(dum1*sinh(kz)*cos(th)   + ...
    4*dum2*sinh(2*kz)*cos(2*th) + ...
    9*dum3*sinh(3*kz)*cos(3*th) + ...
    16*dum4*sinh(4*kz)*cos(4*th) + ...
    25*dum5*sinh(5*kz)*cos(5*th));
  
  zeta = 1/k*((lam                  )*cos(th) + ...
    (lam^2*B22 + lam^4*B24)*cos(2*th) + ...
    (lam^3*B33 + lam^5*B35)*cos(3*th) + ...
    (lam^4*B44            )*cos(4*th) + ...
    (lam^5*B55            )*cos(5*th));

  %% Harmonics of zeta:
  Zeta(1) = 1/k*lam;
  Zeta(2) = 1/k*(lam^2*B22 + lam^4*B24);
  Zeta(3) = 1/k*(lam^3*B33 + lam^5*B35);
  Zeta(4) = 1/k*(lam^4*B44);
  Zeta(5) = 1/k*(lam^5*B55);
  
  %% For zeta > 0:
  % th  = k*x - om*time
  phi1  =  dum1*cosh(kz)*sin(th);
  phi2  =  dum2*cosh(2*kz)*sin(2*th);
  phi3  =  dum3*cosh(3*kz)*sin(3*th);
  phi4  =  dum4*cosh(4*kz)*sin(4*th);
  phi5  =  dum5*cosh(5*kz)*sin(5*th);
  
  cphi1  =  'dum1*cosh(kz)*sin(th);';
  cphi2  =  'dum2*cosh(2*kz)*sin(2*th);';
  cphi3  =  'dum3*cosh(3*kz)*sin(3*th);';
  cphi4  =  'dum4*cosh(4*kz)*sin(4*th);';
  cphi5  =  'dum5*cosh(5*kz)*sin(5*th);';

  %% Rules:
  % d/dt : fac = om; sin  -> -cos;  cos  ->  sin;
  % d/dx : fac = k;  sin  ->  cos;  cos  -> -sin;
  % d/dz : fac = k;  sinh ->  cosh; cosh -> sinh;
  
  kinvec   = {'u','w','ut','ux','uz'};
  irulevec = [ 2   3   2    2    2  ; ...
               0   0   1    2    3   ];
  nkin    = length(kinvec);
  
  rule0  = {'',''};               %  -
  rule1  = {'sin(' ,'cos('} ;     % d/dt
  rule2  = {'sin(' ,'cos('} ;     % d/dx
  rule3  = {'sinh(','cosh('};     % d/dz
  signs  = [  1        1
             -1        1    ; ...
              1       -1    ; ...
              1        1    ];
  facs   = { '1'        '1'  ; ...
             'om'      'om'  ; ...
             'k'        'k'  ; ...
             'k'        'k'  };
  
  %% "Compute" {'u','w','ut','ux','uz'}:
  KIN = zeros(Nz+1,Nt,nkin,nharm);
  for ikin = 1:nkin
    kin = char(kinvec(ikin));
    
    for iharm = 1:nharm
      charm = num2str(iharm);
      eval(['a = cphi' charm ';'])
      
      pm0 = 1;
      for ir = 1:2
        irule = irulevec(ir,ikin);
        eval(['rule = rule' num2str(irule) ';'])
        a1 = char(rule{1});
        a2 = char(rule{2});
        
        ib = strfind(a,a1);
        if (length(ib > 0))
          c    = a1;
          d    = a2;
          pm   = signs(irule+1,1);
          fac  = char(facs(irule+1,1));
        else
          ib = strfind(a,a2);
          c    = a2;
          d    = a1;
          pm   = signs(irule+1,2);
          fac  = char(facs(irule+1,2));
        end
        nr   = length(c);
        ivec = [ib : ib + nr - 1];
        if (length(ivec) > 0)
          a(ivec) = [];
          a = [a(1:ib-1) d a(ib:end)];
        end
        if (irule>0)
          a = [ num2str(iharm) '*' fac '*' a ];
          pm  = pm*pm0;
          pm0 = pm;
        end
      end
      if (pm == -1)
        a = [ num2str(pm) '*' a ];
      end
      %disp(['For z <= 0. ' kin ' :   ' a])
      A{ikin,iharm} = a;
      eval(['dum = ' a ';']);
      KIN(:,:,ikin,iharm) = dum;
    end
  end

   %% Taylor exapansion of {'u','w','ut','ux','uz'}:
  irule = 3;
  rule  = rule3;
  Nder  = 4;
  zm  = max(zeta);
  Nz1 = 120;
  dz  = zm/Nz1;
  z   = [0:Nz1]'*dz;
  K   = zeros(Nz1+1,Nt,nkin);
  
  for ikin = 1:nkin
    kin = char(kinvec(ikin));
    
    %% Zeroth derivative:
    ider = 0;
    for iharm = 1:nharm
      a    = A{ikin,iharm};
      ib   = strfind(a,'kz');
      nr   = length('kz');
      ivec = [ib : ib + nr - 1];
      a(ivec) = [];
      a    = [a(1:ib-1) 'kh' a(ib:end)];
      A0{ikin,iharm,ider+1} = a;
    end
    %['Text 1: ' kin ],A0(ikin,1,:)
    
    %% Derivatives 1-4:
    for ider = 1:Nder
      for iharm = 1:nharm-ider
        a = A0{ikin,iharm,ider};
        
        pm0 = 1;
        a1 = char(rule{1});
        a2 = char(rule{2});
        
        ib = strfind(a,a1);
        if (length(ib > 0))
          c    = a1;
          d    = a2;
          pm   = signs(irule+1,1);
          fac  = char(facs(irule+1,1));
        else
          ib = strfind(a,a2);
          c    = a2;
          d    = a1;
          pm   = signs(irule+1,2);
          fac  = char(facs(irule+1,2));
        end
        nr   = length(c);
        ivec = [ib : ib + nr - 1];
        if (length(ivec) > 0)
          a(ivec) = [];
          a = [a(1:ib-1) d a(ib:end)];
        end
        a = [ num2str(iharm) '*' fac '*' a ];
        pm  = pm*pm0;
        pm0 = pm;
        a = [ num2str(pm) '*' a ];
        
        A0{ikin,iharm,ider+1} = a;
      end
   end
    
    %% Sum up contributions for each z:
    f = zeros(Nz1+1,Nt);
    
    ider = 0;
    for iharm = 1:nharm
      a = A0{ikin,iharm,ider+1};
      eval(['f = f + ones(Nz1+1,1)*' a ])
    end

    for ider = 1:Nder
      for iharm = 1:nharm-ider
        a = A0{ikin,iharm,ider+1};
        str = ['f = f + z.^' num2str(ider) '*' a(1:end-1) '/gamma(' num2str(ider+1) ');'];
        eval(str)
      end
    end
    K(:,:,ikin) = f;
  end
  
  %% Calculate FNV forces:
  u0 = squeeze(K(:,:,1));
  w0 = squeeze(K(:,:,2));
  ut0 = squeeze(K(:,:,3));
  ux0 = squeeze(K(:,:,4));
  uz0 = squeeze(K(:,:,5));
  
   %% Calculate KC5:
  iKCvar = 0;
  if (iKCvar == 1)
    umax = zeros(Nz + Nz1 + 2,1);
    for it = 1:Nt
      if (zeta(it) >=0 )
        ivec0 = find(z <= zeta(it));
        ivec  = [1:Nz+1]';
        uu    = [u(:,it) ; u0(ivec0,it)];
      else
        ivec0 = [];
        ivec  = find(zvec0 <= zeta(it));
        uu    = u(ivec,it);
      end
      ivec1 = [ivec ; ivec0];
      umax(ivec1) = max(umax(ivec1),uu);
    end
    KC5z = umax*T/(2*aa);
  end
  
  tic
  %% Arrange kinematics per time step:
  for it = 1:Nt
    if (zeta(it) >=0 )
      ivec = find(z <= zeta(it));
      uu  = [u(:,it) ; u0(ivec,it)];
      ww  = [w(:,it) ; w0(ivec,it)];
      uut = [ut(:,it) ; ut0(ivec,it)];
      uux = [ux(:,it) ; ux0(ivec,it)];
      uuz = [uz(:,it) ; uz0(ivec,it)];
      zv  = [zvec0 ; z(ivec)];
    else
      ivec = find(zvec0 <= zeta(it));
      uu = u(ivec,it);
      ww = w(ivec,it);
      uut = ut(ivec,it);
      uux = ux(ivec,it);
      uuz = uz(ivec,it);
      zv  = [zvec0(ivec)];
    end

    dzv = diff(zv);
    zvm = 0.5*(zv(1:end-1) + zv(2:end)) + hh1;
    
    um  = 0.5*(uu(1:end-1) + uu(2:end)); 
    wm  = 0.5*(ww(1:end-1) + ww(2:end)); 
    utm = 0.5*(uut(1:end-1) + uut(2:end)); 
    uxm = 0.5*(uux(1:end-1) + uux(2:end)); 
    uzm = 0.5*(uuz(1:end-1) + uuz(2:end)); 
    
    %% KC number dependent a11:
    if (iKCvar == 1)
      [Cm, Cd] = curveFittedCmCd(KC5z);
      a11z     = (Cm - 1)*rho*pi*aa^2;
      a11zm    = 0.5*(a11z(2:end,:) + a11z(1:end-1,:));
      a11      = a11zm(1:length(um));
    end
    
    %% FA:
    nzt(it) = length(zv);
    
    dum1    = 0;%(a11 + fk).*(utm);
    dum2    =          fk.*(um.*uxm);
    dum3    = (a11 + fk).*(wm.*uzm);
    dum     = dum1 + dum2 + dum3;
    FA(it)  = sum(dum.*dzv);
    
    MA(it)  = sum(dum.*zvm.*dzv);
   

    %% FPSI:
    u1    =  KIN(end,it,1,1);
    du1dt =  KIN(end,it,3,1);

    FPSI = fk*4/g*u1.^2.*du1dt;
    
    MPSI = FPSI*hh1;

  end
  
    %% KC dependent drag:
  if (iKCvar == 1)
    Cdm    = 0.5*(Cd(2:end) + Cd(1:end-1));
    whos Cd Cdm um
    dumd   = 0.5*rho*DD*Cdm(1:length(um)).*um.*abs(um);
    FDrag  = iKCvar*sum(dumd.*dzv);
    
    MDrag  = iKCvar*sum(dumd.*zvm.*dzv);
  else
    FDrag  = 0;
    MDrag  = 0;
  end
  
    
%% Plot harmonics
  %% Band-pass filter:
  for iharm = 1:nharm
    flow  = (iharm - 0.4)/T;
    fhigh = (iharm + 0.4)/T;
    nrep = 8;
    
    y   = repmat((FA + FPSI + FDrag),1,nrep)';
    ybp = bpass2(y,dt,flow,fhigh,0,0,'',0.0002);
    Fx(iharm) = max(ybp);

  end
  
  F_FNV = FA +  FPSI + FDrag;
  
end